@@ -16,5 +16,14 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#ifndef MONITOR_H
+#define MONITOR_H
+
+#include <glib.h>
+#include <gio/gio.h>
+
 typedef gboolean (*FileCreationCallback) (const char* crash_file);
-void monitor_directory (const char* directory, FileCreationCallback callback);
+GFileMonitor* monitor_directory (const char* directory, FileCreationCallback callback);
+void unmonitor_directory (GFileMonitor* monitor, FileCreationCallback callback);
+
+#endif /* MONITOR_H */
