@@ -55,8 +55,6 @@
  * can't bring up GNetworkMonitor */
 static gboolean online_state = TRUE;
 
-static GMainLoop* loop = NULL;
-
 /* The URL of the crash database. */
 static char* crash_db_url = NULL;
 
@@ -71,11 +69,16 @@ static char* crash_db_submit_url = NULL;
 /* Username we will run under */
 static const char username[] = "whoopsie";
 
-/* The file descriptor for our instance lock */
+/* The file path and descriptor for our instance lock */
+static const char* lock_path = "/var/lock/whoopsie/lock";
 static int lock_fd = 0;
 
+/* The report directory */
+static const char* report_dir = "/var/crash";
+
 /* Options */
 
+#ifndef TEST
 static gboolean foreground = FALSE;
 
 static GOptionEntry option_entries[] = {
@@ -83,6 +86,7 @@ static GOptionEntry option_entries[] = {
       "Run in the foreground", NULL },
     { NULL }
 };
+#endif
 
 static const char* acceptable_fields[] = {
     "ProblemType",
@@ -133,6 +137,13 @@ static const char* acceptable_fields[] = {
     /* We need the OopsText field to be able to generate a crash signature from
      * KernelOops problems */
     "OopsText",
+    /* We use the UpgradeStatus field to determine the time between upgrading
+     * and the initial error report */
+    "UpgradeStatus",
+    /* We use the InstallationDate and InstallationMedia fields to determine
+     * the time between installing and the initial error report */
+    "InstallationDate",
+    "InstallationMedia",
 
     /* Fields we do not care about: */
 
@@ -163,22 +174,6 @@ is_acceptable_field (const char* field)
     return FALSE;
 }
 
-static void
-parse_arguments (int* argc, char** argv[])
-{
-    GError* err = NULL;
-    GOptionContext* context;
-
-    context = g_option_context_new (NULL);
-    g_option_context_add_main_entries (context, option_entries, NULL);
-    if (!g_option_context_parse (context, argc, argv, &err)) {
-        g_print ("whoopsie: %s\n", err->message);
-        g_error_free (err);
-        exit (EXIT_FAILURE);
-    }
-    g_option_context_free (context);
-}
-
 gboolean
 append_key_value (gpointer key, gpointer value, gpointer bson_string)
 {
@@ -266,6 +261,13 @@ upload_report (const char* message_data, int message_len, struct response_string
 
     g_return_val_if_fail (message_data, -1);
 
+    /* TODO use curl_share for DNS caching. */
+    /* Repeated calls to curl_global_init will have no effect. */
+    if (curl_global_init (CURL_GLOBAL_SSL)) {
+        g_print ("Unable to initialize curl.\n\n");
+        exit (EXIT_FAILURE);
+    }
+
     if ((curl = curl_easy_init ()) == NULL) {
         printf ("Couldn't init curl.\n");
         return FALSE;
@@ -506,14 +508,15 @@ upload_core (const char* uuid, const char* arch, const char* core_data) {
     g_return_val_if_fail (arch, FALSE);
     g_return_val_if_fail (core_data, FALSE);
 
-    if (asprintf (&crash_db_core_url, "%s/%s/submit-core/%s/%s",
-        crash_db_url, uuid, arch, whoopsie_identifier) < 0)
-        g_error ("Unable to allocate memory.");
+    crash_db_core_url = g_strdup_printf ("%s/%s/submit-core/%s/%s",
+                                         crash_db_url, uuid, arch,
+                                         whoopsie_identifier);
 
     /* TODO use CURLOPT_READFUNCTION to transparently compress data with
      * Snappy. */
     if ((curl = curl_easy_init ()) == NULL) {
         printf ("Couldn't init curl.\n");
+        g_free (crash_db_core_url);
         return FALSE;
     }
     init_response_string (&s);
@@ -537,7 +540,7 @@ upload_core (const char* uuid, const char* arch, const char* core_data) {
         curl_easy_strerror (result_code));
     printf ("Response code: %ld\n", response_code);
     curl_easy_cleanup (curl);
-    free (crash_db_core_url);
+    g_free (crash_db_core_url);
     destroy_response_string (&s);
 
     return result_code == CURLE_OK && response_code == 200;
@@ -636,7 +639,7 @@ parse_and_upload_report (const char* crash_file)
 }
 
 void
-process_existing_files (void)
+process_existing_files (const char* report_dir)
 {
     GDir* dir = NULL;
     const gchar* file = NULL;
@@ -644,36 +647,37 @@ process_existing_files (void)
     char* upload_file = NULL;
     char* crash_file = NULL;
 
-    dir = g_dir_open ("/var/crash", 0, NULL);
+    dir = g_dir_open (report_dir, 0, NULL);
     while ((file = g_dir_read_name (dir)) != NULL) {
 
-        upload_file = g_build_filename ("/var/crash", file, NULL);
+        upload_file = g_build_filename (report_dir, file, NULL);
         if (!upload_file)
             continue;
 
         ext = strrchr (upload_file, '.');
         if (ext && strcmp(++ext, "upload") != 0) {
-            free (upload_file);
+            g_free (upload_file);
             continue;
         }
 
         crash_file = change_file_extension (upload_file, ".crash");
         if (!crash_file) {
-            free (upload_file);
+            g_free (upload_file);
             continue;
         }
 
         if (already_handled_report (crash_file)) {
-            if (!mark_handled (crash_file))
-                g_print ("Unable to mark report as seen (%s)\n", crash_file);
+            g_free (upload_file);
+            g_free (crash_file);
+            continue;
         } else if (online_state && parse_and_upload_report (crash_file)) {
             if (!mark_handled (crash_file))
                 g_print ("Unable to mark report as seen (%s)\n", crash_file);
 
         } 
 
-        free (upload_file);
-        free (crash_file);
+        g_free (upload_file);
+        g_free (crash_file);
     }
     g_dir_close (dir);
 }
@@ -716,12 +720,22 @@ void
 exit_if_already_running (void)
 {
     int rc = 0;
-    if (mkdir ("/var/lock/whoopsie", 0755) < 0) {
-        if (errno != EEXIST) {
-            g_print ("Could not create lock directory.\n");
+
+    if (g_getenv ("APPORT_REPORT_DIR")) {
+        /* keep lock file in custom report directory */
+        lock_path = g_build_filename (report_dir, "whoopsie_lock", NULL);
+    } else {
+        /* use system directory */
+        if (mkdir ("/var/lock/whoopsie", 0755) < 0) {
+            if (errno != EEXIST) {
+                g_print ("Could not create lock directory.\n");
+            }
         }
     }
-    lock_fd = open ("/var/lock/whoopsie/lock", O_CREAT | O_RDWR, 0600);
+
+    g_print("Using lock path: %s\n", lock_path);
+
+    lock_fd = open (lock_path, O_CREAT | O_RDWR, 0600);
     rc = flock (lock_fd, LOCK_EX | LOCK_NB);
     if (rc) {
         if (EWOULDBLOCK == errno) {
@@ -733,29 +747,6 @@ exit_if_already_running (void)
     }
 }
 
-static void
-handle_signals (int signo)
-{
-    if (loop)
-        g_main_loop_quit (loop);
-    else
-        exit (0);
-}
-
-static void
-setup_signals (void)
-{
-    struct sigaction action;
-    sigset_t mask;
-
-    sigemptyset (&mask);
-    action.sa_handler = handle_signals;
-    action.sa_mask = mask;
-    action.sa_flags = 0;
-    sigaction (SIGTERM, &action, NULL);
-    sigaction (SIGINT, &action, NULL);
-}
-
 char*
 get_crash_db_url (void)
 {
@@ -777,8 +768,10 @@ drop_privileges (GError** error)
     struct passwd *pw = NULL;
 
     if (getuid () != 0) {
-        g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
-                     "You must be root to run this program.");
+        if (g_getenv ("CRASH_DB_IDENTIFIER") == NULL) {
+            g_set_error (error, g_quark_from_static_string ("whoopsie-quark"), 0,
+                         "You must be root to run this program, or set $CRASH_DB_IDENTIFIER.");
+        }
         return;
     }
     if (!(pw = getpwnam (username))) {
@@ -823,7 +816,7 @@ network_changed (gboolean available)
     online_state = available;
 
     if (online_state)
-        process_existing_files ();
+        process_existing_files (report_dir);
 }
 
 gboolean
@@ -842,11 +835,80 @@ check_online_then_upload (const char* crash_file) {
     return TRUE;
 }
 
+void
+create_directory (void)
+{
+    struct passwd *pw = NULL;
+
+    if (mkdir (report_dir, 0755) < 0) {
+        if (errno != EEXIST) {
+            g_print ("Could not create non-existent directory to monitor (%d): %s.\n", errno, report_dir);
+            exit (EXIT_FAILURE);
+        }
+    } else {
+        /* Only change the permissions if we've just created it */
+        if (!(pw = getpwnam (username))) {
+            g_print ("Could not find user, %s.\n", username);
+            exit (EXIT_FAILURE);
+        }
+        if (chown (report_dir, -1, pw->pw_gid) < 0) {
+            g_print ("Could not change ownership of %s.\n", report_dir);
+            exit (EXIT_FAILURE);
+        }
+        if (chmod (report_dir, 03777) < 0) {
+            g_print ("Could not change permissions on %s.\n", report_dir);
+            exit (EXIT_FAILURE);
+        }
+    }
+}
 #ifndef TEST
+static GMainLoop* loop = NULL;
+
+static void
+parse_arguments (int* argc, char** argv[])
+{
+    GError* err = NULL;
+    GOptionContext* context;
+
+    context = g_option_context_new (NULL);
+    g_option_context_add_main_entries (context, option_entries, NULL);
+    if (!g_option_context_parse (context, argc, argv, &err)) {
+        g_print ("whoopsie: %s\n", err->message);
+        g_error_free (err);
+        exit (EXIT_FAILURE);
+    }
+    g_option_context_free (context);
+}
+
+static void
+handle_signals (int signo)
+{
+    if (loop)
+        g_main_loop_quit (loop);
+    else
+        exit (0);
+}
+
+static void
+setup_signals (void)
+{
+    struct sigaction action;
+    sigset_t mask;
+
+    sigemptyset (&mask);
+    action.sa_handler = handle_signals;
+    action.sa_mask = mask;
+    action.sa_flags = 0;
+    sigaction (SIGTERM, &action, NULL);
+    sigaction (SIGINT, &action, NULL);
+}
+
 int
 main (int argc, char** argv)
 {
     GError* err = NULL;
+    const gchar* env;
+    GFileMonitor* monitor;
 
     setup_signals ();
     parse_arguments (&argc, &argv);
@@ -856,18 +918,30 @@ main (int argc, char** argv)
         exit (EXIT_FAILURE);
     }
 
-    whoopsie_identifier_generate (&whoopsie_identifier, &err);
+    /* environment might change report directory and identifier */
+    env = g_getenv ("APPORT_REPORT_DIR");
+    if (env != NULL)
+        report_dir = env;
+
+    env = g_getenv ("CRASH_DB_IDENTIFIER");
+    if (env != NULL)
+        whoopsie_identifier = g_strdup (env);
+    else
+        whoopsie_identifier_generate (&whoopsie_identifier, &err);
+
     if (err) {
         g_print ("%s\n", err->message);
         g_error_free (err);
         err = NULL;
         crash_db_submit_url = strdup (crash_db_url);
 
-    } else if (asprintf (&crash_db_submit_url, "%s/%s", crash_db_url,
-               whoopsie_identifier) < 0) {
-        g_error ("Unable to allocate memory.");
+    } else {
+        crash_db_submit_url = g_strdup_printf ("%s/%s", crash_db_url,
+                                               whoopsie_identifier);
     }
 
+    create_directory ();
+
     drop_privileges (&err);
     if (err) {
         g_print ("%s\n", err->message);
@@ -879,31 +953,32 @@ main (int argc, char** argv)
     if (!foreground)
         daemonize ();
 
+#if GLIB_MAJOR_VERSION <= 2 && GLIB_MINOR_VERSION < 35
+    /* Deprecated in glib 2.35/2.36. */
     g_type_init ();
+#endif
 
-    /* TODO use curl_share for DNS caching. */
-    if (curl_global_init (CURL_GLOBAL_ALL)) {
-        g_print ("Unable to initialize curl.\n\n");
+    monitor = monitor_directory (report_dir, check_online_then_upload);
+    if (!monitor)
         exit (EXIT_FAILURE);
-    }
 
-    monitor_directory ("/var/crash", check_online_then_upload);
     monitor_connectivity (crash_db_url, network_changed);
-    process_existing_files ();
+    process_existing_files (report_dir);
     g_timeout_add_seconds (PROCESS_OUTSTANDING_TIMEOUT,
-                           (GSourceFunc) process_existing_files, NULL);
+                           (GSourceFunc) process_existing_files, (gpointer) report_dir);
 
     loop = g_main_loop_new (NULL, FALSE);
     g_main_loop_run (loop);
 
-    g_unlink ("/var/lock/whoopsie/lock");
+    unmonitor_directory (monitor, check_online_then_upload);
+    unmonitor_connectivity ();
+
+    g_unlink (lock_path);
     close (lock_fd);
     curl_global_cleanup ();
 
-    if (crash_db_url)
-        free (crash_db_url);
-    if (crash_db_submit_url)
-        free (crash_db_submit_url);
+    g_free (crash_db_url);
+    g_free (crash_db_submit_url);
 	return 0;
 }
 #endif
